from .KnowledgeGraph import SmallKG
from .KnowledgeGraph import KnowledgeGraph

from .DataLoader import load_fb13
from .DataLoader import load_fb15k
from .DataLoader import load_fb15k237
from .DataLoader import load_wn18
from .DataLoader import load_wikidatasets
